#include "testlib.h"
#include <bits/stdc++.h>

using namespace std;

struct node;
typedef node *nodeptr;

auto seed = chrono::high_resolution_clock::now().time_since_epoch().count();
mt19937 mt_rand(seed);
auto randint = uniform_int_distribution<int>(-1e9, 1e9);

struct node {
    nodeptr l = nullptr, r = nullptr;
    int size = 1, y, spaces, last_space;
    char value;

    node(char c) : y(randint(mt_rand)), spaces(c == ' '), value(c) {
        last_space = c == ' ' ? 0 : -1;
    }
};

inline int _size(nodeptr root) {
    return root ? root->size : 0;
}

inline int _spaces(nodeptr root) {
    return root ? root->spaces : 0;
}

inline int _last_space(nodeptr root) {
    return root ? root->last_space : -1;
}

nodeptr recalc(nodeptr root) {
    if (!root) return nullptr;
    root->size = 1 + _size(root->l) + _size(root->r);
    root->spaces = _spaces(root->l) + _spaces(root->r) + (root->value == ' ');
    root->last_space = _last_space(root->l);
    if (root->value == ' ') root->last_space = _size(root->l);
    if (_last_space(root->r) > -1) root->last_space = _size(root->l) + 1 + _last_space(root->r);
    return root;
}

pair<nodeptr, nodeptr> split(nodeptr root, int size) {
    if (!root) return {nullptr, nullptr};
    int lsize = _size(root->l);
    if (lsize < size) {
        auto [l, r] = split(root->r, size - lsize - 1);
        root->r = l;
        return {recalc(root), r};
    } else {
        auto [l, r] = split(root->l, size);
        root->l = r;
        return {l, recalc(root)};
    }
}

nodeptr merge(nodeptr l, nodeptr r) {
    if (!l) return r;
    if (!r) return l;
    if (l->y > r->y) {
        l->r = merge(l->r, r);
        return recalc(l);
    } else {
        r->l = merge(l, r->l);
        return recalc(r);
    }
}

nodeptr search(nodeptr root, int order) {
    int lsize = _size(root->l);
    if (order < lsize) return search(root->l, order);
    if (order > lsize) return search(root->r, order - lsize - 1);
    return root;
}

int space_count(nodeptr root, int order) {
    int lsize = _size(root->l);
    if (order < lsize) return space_count(root->l, order);
    if (order > lsize) return space_count(root->r, order - lsize - 1) + _spaces(root->l) + (root->value == ' ');
    return _spaces(root->l);
}

int main(int argc, char *argv[]) {

    registerGen(argc, argv, 1);

    int len = opt<int>("len"), iw = opt<int>("iw");
    double qp = opt<double>("qp"), sp = opt<double>("sp");
    int n = opt<int>("n"), m = opt<int>("m");

    string s = rnd.next(format("[a-z]{%d}", len));
    auto cuts = rnd.distinct(n - 1, 1, len - 1);
    sort(cuts.begin(), cuts.end());
    cuts.push_back(len);

    vector <string> words(n);
    nodeptr root = nullptr;
    int pos = 0;
    for (int i = 0; i < n; i++) {
        if (i > 0) root = merge(root, new node(' '));
        for (; pos < cuts[i]; pos++) {
            words[i] += s[pos];
            root = merge(root, new node(s[pos]));
        }
    }

    println(n, m);
    println(words.begin(), words.end());
    for (int i = 0; i < m; i++) {
        if (rnd.next() < qp) {
            int index = rnd.wnext(0, root->size - 1, iw);
            if (search(root, index)->value == ' ') {
                index = index < root->size - 1 ? index + 1 : index - 1;
            }
            if (rnd.next() < 0.5) {
                cout << "?1 " << index + 1 << '\n';
            } else {
                auto [l, r] = split(root, index);
                int w = _spaces(l) + 1, p = index - _last_space(l);
                cout << "?2 " << w << " " << p << '\n';
                root = merge(l, r);
            }
        } else {
            int index = rnd.wnext(0, root->size - 1, -iw);
            if (rnd.next() < 0.5) {
                if (rnd.next() < sp) {
                    char c = (char) rnd.next('a', 'z');
                    cout << "+ " << index + 1 << ' ' << c << '\n';
                    auto [l, r] = split(root, index);
                    root = merge(merge(l, new node(c)), r);
                } else {
                    while (index == 0 || search(root, index - 1)->value == ' ' || search(root, index)->value == ' ') {
                        index = (index + 1) % root->size;
                    }
                    cout << "+ " << index + 1 << " _\n";
                    auto [l, r] = split(root, index);
                    root = merge(merge(l, new node(' ')), r);
                }
            } else {
                while (index == 0 || index == (root->size - 1) ||
                        search(root, index)->value != ' ' &&
                               search(root, index - 1)->value == ' ' &&
                               search(root, index + 1)->value == ' ') {
                    index = (index + 1) % root->size;
                }
                cout << "- " << index + 1 << '\n';
                auto [l, tmp] = split(root, index);
                auto [node, r] = split(tmp, 1);
                root = merge(l, r);
            }
        }
    }

}

